package ink.wgink.module.file.service.impl;

import ink.wgink.common.base.DefaultBaseService;
import ink.wgink.exceptions.FileException;
import ink.wgink.exceptions.base.SystemException;
import ink.wgink.module.file.dao.IFileDao;
import ink.wgink.module.file.enums.UploadTypeEnum;
import ink.wgink.module.file.service.IFileService;
import ink.wgink.module.file.service.IMinIoFileService;
import ink.wgink.pojo.pos.FilePO;
import ink.wgink.properties.FileMinIoProperties;
import ink.wgink.properties.FileProperties;
import ink.wgink.util.UUIDUtil;
import ink.wgink.util.request.StaticResourceRequestUtil;
import io.minio.MinioClient;
import io.minio.PutObjectOptions;
import io.minio.errors.*;
import org.apache.commons.lang3.StringUtils;
import org.joda.time.DateTime;
import org.joda.time.DateTimeZone;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;

import javax.annotation.PostConstruct;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URLEncoder;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.util.List;
import java.util.Map;

/**
 * @ClassName: MinIOServiceImpl
 * @Description: minIO
 * @Author: wanggeng
 * @Date: 2021/10/17 10:16 上午
 * @Version: 1.0
 */
@Service
public class MinIoFileServiceImpl extends DefaultBaseService implements IMinIoFileService {

    @Autowired
    private FileProperties fileProperties;
    @Autowired
    private IFileDao fileDao;
    private FileMinIoProperties minIoProperties;
    private MinioClient minioClient;


    @PostConstruct
    public void init() throws InvalidPortException, InvalidEndpointException {
        if (!fileProperties.getUseMinIo()) {
            return;
        }
        minIoProperties = fileProperties.getMinIo();
        minioClient = new MinioClient(minIoProperties.getEndpoint(), minIoProperties.getAccessKey(), minIoProperties.getSecretKey(), minIoProperties.getSecure());
    }

    @Override
    public void deleteFile(List<FilePO> filePOs) {

    }

    @Override
    public void uploadFile(String token, MultipartFile uploadFile, UploadTypeEnum uploadTypeEnum, Map<String, Object> params) {
        uploadFile(token, uploadFile, uploadTypeEnum, params, false);
    }

    @Override
    public void uploadFileByUserId(String userId, MultipartFile uploadFile, UploadTypeEnum uploadTypeEnum, Map<String, Object> params) {
        uploadFile(userId, uploadFile, uploadTypeEnum, params, true);
    }

    /**
     * 上传文件
     *
     * @param tokenOrUserId
     * @param uploadFile
     * @param uploadTypeEnum
     * @param params
     * @param useUserId      是否使用外来userId
     */
    private void uploadFile(String tokenOrUserId, MultipartFile uploadFile, UploadTypeEnum uploadTypeEnum, Map<String, Object> params, boolean useUserId) {
        String fileId = UUIDUtil.getUUID();
        params.put("fileId", fileId);

        String fileName = uploadFile.getOriginalFilename();
        // 文件大小
        long fileSize = uploadFile.getSize();
        // 文件类型
        String fileType = getFileType(fileName);
        String uploadFileName = fileId + "." + fileType;
        String bucketName = getBucketName(uploadTypeEnum);
        // 上传文件
        upload(uploadFileName, fileSize, bucketName, uploadFile);
        if (useUserId) {
            saveFileByUserId(tokenOrUserId, fileId, fileName, fileType, fileSize, bucketName);
        } else {
            saveFile(tokenOrUserId, fileId, fileName, fileType, fileSize, bucketName);
        }
    }

    @Override
    public void downLoadFile(HttpServletRequest request, HttpServletResponse response, FilePO filePO, boolean isOpen, boolean canRange) {
        // minIo中文件名
        String objectName = filePO.getFileId() + "." + filePO.getFileType();
        InputStream inputStream = null;
        try (OutputStream outputStream = response.getOutputStream();) {
            response.setHeader("Content-Length", filePO.getFileSize());
            response.setContentType(StaticResourceRequestUtil.getContentType(filePO.getFileType()));
            if (!isOpen) {
                // 下载
                response.setHeader("Content-Disposition", "attachment;fileName=" + URLEncoder.encode(filePO.getFileName(), "UTF-8"));
                inputStream = minioClient.getObject(filePO.getFileUrl(), objectName);
            } else {
                // 直接打开
                response.setHeader("Content-Disposition", "inline;fileName=" + URLEncoder.encode(filePO.getFileName(), "UTF-8"));
                // 如果是图片资源，开启缓存
                if (isImageFile(filePO.getFileType())) {
                    // 如果存在校验修改时间且未做修改，返回304，使用本地资源
                    String ifModifiedSince = request.getHeader("If-Modified-Since");
                    if (StringUtils.isNotBlank(ifModifiedSince) && StringUtils.equalsIgnoreCase(ifModifiedSince, filePO.getGmtModified())) {
                        response.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
                        return;
                    }
                    // 缓存有效时间为7天
                    response.setHeader("Expires", DateTime.now().plusDays(7).toDateTime(DateTimeZone.forID("GMT")).toString());
                    // 一小时之内不发送新请求
                    response.setHeader("max-age", "3600");
                    response.setHeader("Last-Modified", filePO.getGmtModified());
                }
                String rangeString = null;
                if (canRange && request != null) {
                    rangeString = request.getHeader("Range");
                    LOG.debug("range: {}", rangeString);
                }
                long contentLength = Long.valueOf(filePO.getFileSize());
                long startRange = 0;
                long endRange = contentLength - 1;
                if (!StringUtils.isBlank(rangeString)) {
                    if (!isOpen) {
                        response.setContentType("multipart/byteranges");
                    }
                    String[] rangeArray = rangeString.substring(rangeString.indexOf("=") + 1).split("-");
                    startRange = Long.valueOf(rangeArray[0]);
                    if (rangeArray.length > 1) {
                        endRange = Long.valueOf(rangeArray[1]);
                    }
                    setRangeHeader(startRange, endRange, response, filePO.getFileId(), contentLength);
                    inputStream = minioClient.getObject(filePO.getFileUrl(), objectName, startRange, endRange);
                } else {
                    inputStream = minioClient.getObject(filePO.getFileUrl(), objectName);
                }
            }
            byte[] readBuf = new byte[IFileService.INPUT_STREAM_SIZE];
            for (int length = 0; (length = inputStream.read(readBuf)) > 0; ) {
                outputStream.write(readBuf, 0, length);
            }
            outputStream.flush();
            inputStream.close();
        } catch (IOException | ErrorResponseException | InsufficientDataException | InternalException | InvalidBucketNameException | InvalidKeyException | InvalidResponseException | NoSuchAlgorithmException | XmlParserException e) {
            throw new FileException("文件输出异常", e);
        } finally {
            if (inputStream != null) {
                try {
                    inputStream.close();
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
    }


    /**
     * 上传文件到minIo
     *
     * @param uploadFileName
     * @param fileSize
     * @param bucketName
     * @param uploadFile
     */
    private void upload(String uploadFileName, long fileSize, String bucketName, MultipartFile uploadFile) {
        try (InputStream inputStream = uploadFile.getInputStream()) {
            // 创建桶
            if (!minioClient.bucketExists(bucketName)) {
                minioClient.makeBucket(bucketName);
            }
            PutObjectOptions putObjectOptions = new PutObjectOptions(fileSize, PutObjectOptions.MAX_PART_SIZE);
            minioClient.putObject(bucketName, uploadFileName, inputStream, putObjectOptions);
        } catch (MinioException | InvalidKeyException | IOException | NoSuchAlgorithmException e) {
            throw new SystemException(e);
        }
    }

    /**
     * 保存文件
     *
     * @param token
     * @param fileId
     * @param fileName
     * @param fileType
     * @param fileSize
     * @param bucketName 桶名称
     */
    private void saveFile(String token, String fileId, String fileName, String fileType, long fileSize, String bucketName) {
        Map<String, Object> params = getFileParams(fileId, fileName, fileType, fileSize, bucketName);
        if (StringUtils.isBlank(token)) {
            setSaveInfo(params);
        } else {
            setAppSaveInfo(token, params);
        }
        fileDao.save(params);
    }

    /**
     * 保存用户
     *
     * @param userId     用户ID
     * @param fileId
     * @param fileName
     * @param fileType
     * @param fileSize
     * @param bucketName
     */
    private void saveFileByUserId(String userId, String fileId, String fileName, String fileType, long fileSize, String bucketName) {
        Map<String, Object> params = getFileParams(fileId, fileName, fileType, fileSize, bucketName);
        setSaveInfoByUserId(params, userId);
        fileDao.save(params);
    }

    private Map<String, Object> getFileParams(String fileId, String fileName, String fileType, long fileSize, String bucketName) {
        Map<String, Object> params = getHashMap(2);
        params.put("fileId", fileId);
        params.put("fileName", fileName);
        params.put("filePath", null);
        params.put("fileUrl", bucketName);
        params.put("fileType", fileType);
        params.put("fileSize", fileSize);
        params.put("fileSummary", "minio");
        params.put("isBack", 1);
        return params;
    }

    /**
     * 获取文件类型
     *
     * @param fileName
     * @return
     */
    private String getFileType(String fileName) {
        String[] names = fileName.split("\\.");
        if (names != null) {
            return names[names.length - 1].toLowerCase();
        }
        return "";
    }

    /**
     * 获取桶文件名
     *
     * @param uploadTypeEnum 上传类型
     * @return
     */
    private String getBucketName(UploadTypeEnum uploadTypeEnum) {
        if (uploadTypeEnum.getValue() == UploadTypeEnum.IMAGE.getValue()) {
            return "images";
        } else if (uploadTypeEnum.getValue() == UploadTypeEnum.VIDEO.getValue()) {
            return "videos";
        } else if (uploadTypeEnum.getValue() == UploadTypeEnum.AUDIO.getValue()) {
            return "audios";
        }
        return "files";
    }

    /**
     * 是否是图片
     *
     * @param fileType
     * @return
     */
    private boolean isImageFile(String fileType) {
        if (StringUtils.equalsIgnoreCase("gif", fileType)) {
            return false;
        }
        String imageTypes = fileProperties.getImageTypes();
        for (String imageType : imageTypes.split(",")) {
            if (StringUtils.equalsIgnoreCase(fileType, imageType)) {
                return true;
            }
        }
        return false;
    }

    /**
     * 处理视频流问题
     *
     * @param startRange
     * @param endRange
     * @param response
     * @param fileId
     * @param contentLength
     */
    private void setRangeHeader(long startRange, long endRange, HttpServletResponse response, String fileId, long contentLength) {
        // 这里不设置，会出现第一次加载很慢的情况
        response.setHeader("Content-Length", String.valueOf(endRange - startRange + 1));
        response.setHeader("Content-Range", String.format("bytes %d-%d/%d", startRange, endRange, contentLength));
        response.setHeader("Accept-Ranges", "bytes");
        response.setHeader("Etag", fileId);
        response.setStatus(206);
    }
}
