package ink.wgink.module.instantmessage.websocket.manager;

import com.alibaba.fastjson.JSONObject;
import ink.wgink.module.instantmessage.websocket.exception.SessionException;
import ink.wgink.module.instantmessage.websocket.pojo.WebSocketClientMessage;
import ink.wgink.module.instantmessage.websocket.pojo.WebSocketSession;
import io.netty.channel.Channel;
import io.netty.channel.group.ChannelGroup;
import io.netty.channel.group.DefaultChannelGroup;
import io.netty.handler.codec.http.websocketx.TextWebSocketFrame;
import io.netty.util.concurrent.GlobalEventExecutor;
import org.apache.commons.lang3.StringUtils;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

/**
 * @ClassName: WebSocketChannelManager
 * @Description: WebSocket通道管理器
 * @Author: wanggeng
 * @Date: 2021/9/11 11:56 上午
 * @Version: 1.0
 */
public class WebSocketChannelManager {
    public static final String FORM_SYSTEM = "SYSTEM";
    private static final WebSocketChannelManager webSocketChannelManager = WebSocketChannelManagerBuilder.webSocketChannelManager;
    private ChannelGroup globalGroup = new DefaultChannelGroup(GlobalEventExecutor.INSTANCE);
    private ConcurrentMap<String, List<WebSocketSession>> webSocketUserMap = new ConcurrentHashMap<>();

    private WebSocketChannelManager() {
    }

    public static WebSocketChannelManager getInstance() {
        return webSocketChannelManager;
    }

    /**
     * 初始化会话队列
     *
     * @param userId     用户ID
     * @param sessionId  会话ID
     * @param clientName 客户端名称
     */
    public void init(String userId, String sessionId, String clientName) {
        WebSocketSession onlineUser = getOnlineUser(userId, clientName);
        if (onlineUser == null) {
            onlineUser = new WebSocketSession();
            onlineUser.setUserId(userId);
            onlineUser.setClientName(clientName);
            // 更新到session队列中
            List<WebSocketSession> webSocketSessions = listOnlineUser(userId);
            webSocketSessions.add(onlineUser);
            webSocketUserMap.put(userId, webSocketSessions);
        }
        // 更新最后时间
        onlineUser.setSessionId(sessionId);
        onlineUser.setUpdateTime(System.currentTimeMillis());
    }

    /**
     * 添加通道，绑定通道与会话
     *
     * @param sessionId
     * @param userId
     * @param channel
     * @throws SessionException
     */
    public void addChannel(String sessionId, String userId, Channel channel) throws SessionException {
        WebSocketSession onlineUser = getOnlineUserBySessionId(sessionId);
        if (onlineUser == null) {
            throw new SessionException("无效会话，请登录");
        }
        if (!StringUtils.equals(userId, onlineUser.getUserId())) {
            throw new SessionException("from错误");
        }
        // 更新会话通道
        onlineUser.setChannel(channel);
        onlineUser.setChannelId(channel.id().asLongText());
        globalGroup.add(channel);
    }

    public void removeChannel(Channel channel) {
        globalGroup.remove(channel);
        for (Map.Entry<String, List<WebSocketSession>> kv : webSocketUserMap.entrySet()) {
            List<WebSocketSession> webSocketSessions = kv.getValue();
            for (int i = 0; i < webSocketSessions.size(); i++) {
                WebSocketSession webSocketSession = webSocketSessions.get(i);
                if (StringUtils.equals(webSocketSession.getChannelId(), channel.id().asLongText())) {
                    webSocketSessions.remove(i);
                    return;
                }
            }
        }
    }

    /**
     * 获取在线用户客户端列表
     *
     * @param userId 用户ID
     * @return
     */
    public List<WebSocketSession> listOnlineUser(String userId) {
        if (StringUtils.isBlank(userId)) {
            return new ArrayList<>();
        }
        List<WebSocketSession> webSocketSessions = webSocketUserMap.get(userId);
        if (webSocketSessions == null) {
            return new ArrayList<>();
        }
        return webSocketSessions;
    }


    /**
     * 获取在线用户
     *
     * @param userId     用户ID
     * @param clientName 客户端名称
     * @return
     */
    public WebSocketSession getOnlineUser(String userId, String clientName) {
        if (StringUtils.isBlank(userId) || StringUtils.isBlank(clientName)) {
            return null;
        }
        List<WebSocketSession> webSocketSessions = listOnlineUser(userId);
        if (webSocketSessions.isEmpty()) {
            return null;
        }
        for (WebSocketSession webSocketSession : webSocketSessions) {
            if (StringUtils.equals(clientName, webSocketSession.getClientName())) {
                return webSocketSession;
            }
        }
        return null;
    }

    /**
     * 获取在线用户客户端列表
     *
     * @param userIds 用户ID列表
     * @return
     */
    public List<WebSocketSession> listOnlineUser(List<String> userIds) {
        if (userIds == null || userIds.isEmpty()) {
            return new ArrayList<>();
        }
        List<WebSocketSession> webSocketSessions = new ArrayList<>();
        for (String userId : userIds) {
            List<WebSocketSession> users = webSocketUserMap.get(userId);
            if (users != null) {
                webSocketSessions.addAll(users);
            }
        }
        return webSocketSessions;
    }

    /**
     * 获取在线用户客户端列表
     *
     * @param userIds    用户ID列表
     * @param clientName 客户端名称
     * @return
     */
    public List<WebSocketSession> listOnlineUser(List<String> userIds, String clientName) {
        if (userIds == null || userIds.isEmpty() || StringUtils.isBlank(clientName)) {
            return new ArrayList<>();
        }
        List<WebSocketSession> webSocketSessions = new ArrayList<>();
        for (String userId : userIds) {
            List<WebSocketSession> users = webSocketUserMap.get(userId);
            if (users == null) {
                continue;
            }
            for (WebSocketSession user : users) {
                if (StringUtils.equals(clientName, user.getClientName())) {
                    webSocketSessions.add(user);
                }
            }
        }
        return webSocketSessions;
    }

    /**
     * 通过会话获取用户
     *
     * @param sessionId 会话ID
     * @return
     */
    private WebSocketSession getOnlineUserBySessionId(String sessionId) {
        for (Map.Entry<String, List<WebSocketSession>> kv : webSocketUserMap.entrySet()) {
            for (WebSocketSession webSocketSession : kv.getValue()) {
                if (StringUtils.equals(sessionId, webSocketSession.getSessionId())) {
                    return webSocketSession;
                }
            }
        }
        return null;
    }

    /**
     * 通过通道获取用户
     *
     * @param channelId 通道ID
     * @return
     */
    private WebSocketSession getOnlineUserByChannelId(String channelId) {
        for (Map.Entry<String, List<WebSocketSession>> kv : webSocketUserMap.entrySet()) {
            for (WebSocketSession webSocketSession : kv.getValue()) {
                if (StringUtils.equals(channelId, webSocketSession.getChannelId())) {
                    return webSocketSession;
                }
            }
        }
        return null;
    }

    /**
     * 全部发
     *
     * @param tws
     */
    public void sendAll(TextWebSocketFrame tws) {
        globalGroup.writeAndFlush(tws);
    }

    /**
     * 单发
     *
     * @param toChannel              接收人通道
     * @param webSocketClientMessage 发送内容
     */
    public void sendText(Channel toChannel, WebSocketClientMessage webSocketClientMessage) {
        if (toChannel == null || !toChannel.isOpen() || !toChannel.isActive()) {
            return;
        }
        TextWebSocketFrame textWebSocketFrame = new TextWebSocketFrame(JSONObject.toJSONString(webSocketClientMessage));
        toChannel.writeAndFlush(textWebSocketFrame);
    }

    /**
     * 群发
     *
     * @param channels
     * @param webSocketClientMessage
     */
    public void sendGroupText(List<Channel> channels, WebSocketClientMessage webSocketClientMessage) {
        for (Channel channel : channels) {
            sendText(channel, webSocketClientMessage);
        }
    }

    private static class WebSocketChannelManagerBuilder {
        public static final WebSocketChannelManager webSocketChannelManager = new WebSocketChannelManager();
    }

}
